#include "fcgi_stdio.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <openssl/sha.h>
#include "libpq-fe.h"
#include "inno-lib.h"
#include "reporting_config.h"
#include "pg_report.h"

#define PG_CONN_STR_REMOTE      "host = %s port = 5432 dbname = innovaphone-reporting user = innovaphone-reporting password = '%s' sslmode = require"
#define XSL_REPORT              "<?xml-stylesheet type='text/xsl' href='call_list.xsl'?>"
#define XSL_CALLBACK            "<?xml-stylesheet type='text/xsl' href='callback.xsl'?>"
#define XSL_MISSED              "<?xml-stylesheet type='text/xsl' href='missed_calls.xsl'?>"
#define XSL_CLEAR               "<?xml-stylesheet type='text/xsl' href='call_list.xsl'?>"
#define TAG_AUTH                "authenticate"
#define TAG_CALLBACK            "callback"
#define TAG_MISSED              "missed-calls"
#define TAG_CLEAR               "clear"
#define QUERY_CALLBACK          "UPDATE cdr_properties SET cdrp_callback_dn='%s', cdrp_callback_h323='%s', cdrp_callback_e164='%s', cdrp_callback_time=date_trunc('second', localtimestamp) WHERE cdrp_conf='%s'"
#define QUERY_MISSED            "SELECT count(DISTINCT cdrp.cdrp_conf) AS count FROM cdrs cdr JOIN cdr_users cu ON cdr.cdr_user_id=cu.id JOIN cdr_properties cdrp ON cdr.id=cdrp.cdr_id WHERE cu.cn='%s' AND cu.%s='%s' AND cdr.local_stamp > cu.missed_calls_time AND cdr.transferred=0 AND cdr.forwarded=0 AND cdrp.cdrp_call_flow LIKE '%%{setup-to,%%,%%,%%}%%' AND cdrp.cdrp_call_flow NOT LIKE '%%{conn%%,%%,%%,%%}%%' AND cdrp.cause_code!=26"
#define QUERY_CLEAR_UPDATE      "UPDATE cdr_users SET clear_report_time=date_trunc('second', localtimestamp) WHERE cn='%s' AND %s='%s'"

void cleanup(CURL * conn, char * action, char * cn, char * sys, char * pbx, char * challenge, char * auth, config_t *reporting_cfg)
{
    curl_easy_cleanup(conn);
    if(action) free(action);
    if(cn) free(cn);
    if(sys) free(sys);
    if(pbx) free(pbx);
    if(challenge) free(challenge);
    if(auth) free(auth);
    if(reporting_cfg) cleanup_config(reporting_cfg);
}

void send_xml(char * attrib_name, char * attrib_value, int attrib_value_len, const char * tag, const char * xsl, FCGX_Request *cgi, int *bytes_sent, int *output_len)
{
    if(attrib_name && !strcasecmp(attrib_name, "error")) {
        do_log(LOG_REPORTING, "", "%s", attrib_value);
    }
    static char * xml_start = "<?xml version=\"1.0\" encoding=\"utf-8\"?>";
    int xml_len = strlen(xml_start) + strlen(xsl) + strlen(tag) + 4;
    char * xml_attrib = NULL;
    if(attrib_value_len > 0) {
        xml_attrib = xml_entities(attrib_value);
        xml_len += strlen(attrib_name) + strlen(xml_attrib) + 4;
    }
    char * xml = malloc(xml_len);
    int len = sprintf(xml, "%s%s<%s", xml_start, xsl, tag);
    if(attrib_value_len > 0) {
        len += sprintf(&xml[len], " %s=\"%s\"", attrib_name, xml_attrib);
        free(xml_attrib);
    }
    len += sprintf(&xml[len], "/>");
    char xml_len_str[11];
    snprintf(xml_len_str, sizeof(xml_len_str), "%d", len);
    *bytes_sent = FCGX_PutS(EXPIRES_HEADER, cgi->out);
    *bytes_sent += FCGX_PutS(CONTENT_TYPE_XML, cgi->out);
    *bytes_sent += FCGX_PutS(CONTENT_LENGTH, cgi->out);
    *bytes_sent += FCGX_PutS(xml_len_str, cgi->out);
    *bytes_sent += FCGX_PutS("\r\n\r\n", cgi->out);
    *bytes_sent += FCGX_PutS(xml, cgi->out);
    *output_len = len;
    *output_len += strlen(EXPIRES_HEADER);
    *output_len += strlen(CONTENT_TYPE_XML);
    *output_len += strlen(CONTENT_LENGTH);
    *output_len += strlen(xml_len_str);
    *output_len += 4;
    free(xml);
}

char * read_file(char * file)
{
    FILE * fp = fopen(file, "r");
    if(fp) {
        fseek(fp, 0L, SEEK_END);
        int file_size = ftell(fp);
        fseek(fp, 0L, SEEK_SET);
        char * content = malloc(file_size + 1);
        fread(content, 1, file_size, fp);
        fclose(fp);
        content[file_size] = 0;
        return content;
    }
    return NULL;
}

void cleanup_callback(char * dn, char * name, char * number, char * conf, char * query) 
{
    if(dn) free(dn);
    if(name) free(name);
    if(number) free(number);
    if(conf) free(conf);
    if(query) free(query);
}

PGconn * connect_to_db(char * conn_str, char * type, char * action, char * log_str, int log_size, int * log_len) 
{
    PGconn * conn = PQconnectdb(conn_str);
    if (PQstatus(conn) != CONNECTION_OK) {
        *log_len = snprintf(log_str, log_size, "mypbx (%s): connection to %s database failed: %s", action, type, PQerrorMessage(conn));
        PQfinish(conn);
        return NULL;
    }

    if (PQsetClientEncoding(conn, "UTF-8") != 0) {
        *log_len = snprintf(log_str, log_size, "mypbx (%s): setting %s client encoding failed: %s", action, type, PQerrorMessage(conn));
        PQfinish(conn);
        return NULL;
    }
    return conn;
}

int missed_calls(CURL * conn, char * query_string, char * cn, char * sys, char * pbx, FCGX_Request *cgi, const char * xsl, int *bytes_sent, int *output_len, config_t *reporting_cfg)
{
    char log_str[MAX_TRACE_LENGTH];
    int log_len = 0;
    PGresult * res = NULL;
    PGconn *pg_conn = NULL;
    char * query = NULL;
    char * missed = NULL;
    char * userpbx = sys;
    if(!strcmp(reporting_cfg->pbx->identifier, "pbx")) {
        userpbx = pbx;
    }

    pg_conn = connect_to_db(PG_CONN_STR, "local", ACTION_MISSED, log_str, MAX_TRACE_LENGTH, &log_len);
    if(!pg_conn) {
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_MISSED, xsl, cgi, bytes_sent, output_len);
        return ERROR_CONNECTING_TO_DB;
    }
    query = malloc(strlen(QUERY_MISSED) + strlen(cn) + strlen(userpbx) + strlen(reporting_cfg->pbx->identifier) + 1);
    sprintf(query, QUERY_MISSED, cn, reporting_cfg->pbx->identifier, userpbx);

    res = PQexec(pg_conn, query);
    free(query);
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): fetching missed calls count failed: %s", ACTION_MISSED, PQerrorMessage(pg_conn));
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_MISSED, xsl, cgi, bytes_sent, output_len);
        PQclear(res);
        PQfinish(pg_conn);
        return ERROR_MYPBX_FETCHING_MISSED_CALLS;
    }
    missed = PQgetvalue(res, 0, 0);
    send_xml("count", missed, strlen(missed), TAG_MISSED, xsl, cgi, bytes_sent, output_len);
    PQclear(res);
    PQfinish(pg_conn);
    return 0;
}

int clear(CURL * conn, char * query_string, char * cn, char * sys, char * pbx, FCGX_Request *cgi, const char * xsl, int *bytes_sent, int *output_len, config_t *reporting_cfg)
{
    char log_str[MAX_TRACE_LENGTH];
    int log_len = 0;
    PGresult * res = NULL;
    PGconn *pg_conn = NULL;
    char * query = NULL;
    char * userpbx = sys;
    if(!strcmp(reporting_cfg->pbx->identifier, "pbx")) {
        userpbx = pbx;
    }
    int error = 0;

    if(check_db_restore()) {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): db being restored", ACTION_MISSED);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_MISSED, xsl, cgi, bytes_sent, output_len);
        return ERROR_DB_BEING_RESTORED;
    }

    pg_conn = connect_to_db(PG_CONN_STR, "local", ACTION_MISSED, log_str, MAX_TRACE_LENGTH, &log_len);
    if(!pg_conn) {
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_MISSED, xsl, cgi, bytes_sent, output_len);
        return ERROR_CONNECTING_TO_DB;
    }

    query = malloc(strlen(QUERY_CLEAR_UPDATE) + strlen(cn) + strlen(userpbx) + strlen(reporting_cfg->pbx->identifier) + 1);
    sprintf(query, QUERY_CLEAR_UPDATE, cn, reporting_cfg->pbx->identifier, userpbx);
    res = PQexec(pg_conn, query);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): updating clear report time for cn=%s %s=%s failed: %s", ACTION_CLEAR, cn, reporting_cfg->pbx->identifier, userpbx, PQerrorMessage(pg_conn));
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_CLEAR, xsl, cgi, bytes_sent, output_len);
        error = ERROR_MYPBX_UPDATING_CLEAR_REPORT;
    }
    PQclear(res);
    PQfinish(pg_conn);
    free(query);
    if(error == 0) {
        send_xml(NULL, NULL, 0, TAG_CLEAR, xsl, cgi, bytes_sent, output_len);
    }
    return error;
}

int callback(CURL * conn, char * query_string, FCGX_Request *cgi, const char * xsl, int *bytes_sent, int *output_len, config_t *reporting_cfg)
{
    char log_str[MAX_TRACE_LENGTH];
    int log_len = 0;
    char * dn = NULL;
    char * name = NULL;
    char * number = NULL;
    char * conf = NULL;
    char * query = NULL;
    PGresult * res = NULL;
    PGconn *pg_conn = NULL;
    PGconn *pg_remote = NULL;
    int query_len = 0;
    char * conn_str_remote = NULL;
    
    if(check_db_restore()) {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): db being restored", ACTION_CALLBACK);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_CALLBACK, xsl, cgi, bytes_sent, output_len);
        return ERROR_DB_BEING_RESTORED;
    }
    
    dn = get_url_param(conn, query_string, "dn", "mypbx", LOG_REPORTING);
    name = get_url_param(conn, query_string, "name", "mypbx", LOG_REPORTING);
    number = get_url_param(conn, query_string, "number", "mypbx", LOG_REPORTING);
    conf = get_url_param(conn, query_string, "conf", "mypbx", LOG_REPORTING);
    if(((!dn || !*dn) && (!name || !*name) && (!number || !*number)) || !conf || !*conf) {
        cleanup_callback(dn, name, number, conf, query);
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): not all necessary callback parameters set", ACTION_CALLBACK);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_CALLBACK, xsl, cgi, bytes_sent, output_len);
        return ERROR_MYPBX_CB_PARAMS_NOT_SET;
    }
    
    pg_conn = connect_to_db(PG_CONN_STR, "local", ACTION_CALLBACK, log_str, MAX_TRACE_LENGTH, &log_len);
    if (!pg_conn) {
        cleanup_callback(dn, name, number, conf, query);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_CALLBACK, xsl, cgi, bytes_sent, output_len);
        return ERROR_CONNECTING_TO_DB;
    }
    
    if(!dn) dn = str_malloc("");
    if(!name) name = str_malloc("");
    if(!number) number = str_malloc("");

    query_len = strlen(QUERY_CALLBACK) + strlen(dn) + strlen(name) + strlen(number) + strlen(conf) + 1;
    query = malloc(query_len);
    snprintf(query, query_len, QUERY_CALLBACK, dn, name, number, conf);

    res = PQexec(pg_conn, query);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        PQclear(res);
        cleanup_callback(dn, name, number, conf, query);
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): updating callback info failed: %s", ACTION_CALLBACK, PQerrorMessage(pg_conn));
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_CALLBACK, xsl, cgi, bytes_sent, output_len);
        PQfinish(pg_conn);
        return ERROR_MYPBX_CB_UPDATE_INFO_FAILED;
    }
    PQclear(res);
    PQfinish(pg_conn);
    // update callback info on the master/standby db
    if(!reporting_cfg->error && reporting_cfg->replication->mode && 
        (!strcasecmp(reporting_cfg->replication->mode, "standby") || !strcasecmp(reporting_cfg->replication->mode, "master")) &&
        reporting_cfg->replication->server && reporting_cfg->replication->password) {
        conn_str_remote = malloc(strlen(PG_CONN_STR_REMOTE) + strlen(reporting_cfg->replication->server) + strlen(reporting_cfg->replication->password) + 1);
        sprintf(conn_str_remote, PG_CONN_STR_REMOTE, reporting_cfg->replication->server, reporting_cfg->replication->password);
        pg_remote = connect_to_db(conn_str_remote, "remote", ACTION_CALLBACK, log_str, MAX_TRACE_LENGTH, &log_len);
        free(conn_str_remote);
        if(pg_remote) {
            res = PQexec(pg_remote, query);
            if(PQresultStatus(res) != PGRES_COMMAND_OK) {
                do_log(LOG_REPORTING, "", "mypbx (%s): updating callback info on remote server failed: %s", ACTION_CALLBACK, PQerrorMessage(pg_remote));
            }
            PQclear(res);
            PQfinish(pg_remote);
        }
        else {
            do_log(LOG_REPORTING, "", "%s", log_str);
        }
    }
    send_xml(NULL, NULL, 0, TAG_CALLBACK, xsl, cgi, bytes_sent, output_len);
    cleanup_callback(dn, name, number, conf, query);
    return 0;
}

int report(CURL * conn, char * query_string, char * cn, char * sys, char * pbx, FCGX_Request *cgi, char * doc_root, const char * xsl, int *bytes_sent, int *output_len, config_t *reporting_cfg)
{
    char log_str[MAX_TRACE_LENGTH];
    int log_len = 0;
    int query_len = 0;
    char tmp_template[128];
    char tmp_query_string[140];
    char * limit = NULL;
    char * offset = NULL;
    const char * date_format = NULL;
    char * lang_str = NULL;
    FILE * fp_tempfile = NULL;
    int tempfile = 0;
    char * query = NULL;
    int i = 0;
    char cols[11];
    int res = 0, lang;
    char * date_time = NULL;
    time_t ti = time(NULL) - 31*24*60*60;   // substract 31 days from current time
    struct tm * t = localtime(&ti);
    char old_time[11];                      // holds time in format YYYY-MM-DD
    char * userpbx = sys;
    if(!strcmp(reporting_cfg->pbx->identifier, "pbx")) {
        userpbx = pbx;
    }
    
    snprintf(tmp_template, sizeof(tmp_template), "%s/apps/innovaphone-reporting/tmp/mypbxcfgXXXXXX", doc_root);
    tempfile = mkstemp(tmp_template);
    if(!tempfile || !(fp_tempfile = fdopen(tempfile, "w+"))) {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): creating temporary config file '%s' failed", ACTION_REPORT, tmp_template);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
        return ERROR_MYPBX_CREATING_TMP_FILE;
    }
    limit = get_url_param(conn, query_string, "count", "mypbx", LOG_REPORTING);
    offset = get_url_param(conn, query_string, "offset", "mypbx", LOG_REPORTING);
    lang_str = get_url_param(conn, query_string, "lang", "mypbx", LOG_REPORTING);
    lang = lookup_language(lang_str);
    date_format = process_string_table[TXT_DATE_FORMAT].lang[lang];
    if(!limit) limit = str_malloc("100");
    if(!offset) offset = str_malloc("");
    if(date_format && strlen(date_format) < 10) {
        date_time = malloc(30);
        *date_time = 0;
        const char * temp = date_format;
        while(*temp) {
            switch(*temp) {
                case 'Y': strcat(date_time, "YYYY"); break;
                case 'D': strcat(date_time, "DD"); break;
                case 'M': strcat(date_time, "MM"); break;
                default: strncat(date_time, temp, 1); break;
            }
            temp++;
        }
        strcat(date_time, " HH24:MI:SS");
    }
    else {
        date_time = str_malloc("DD/MM/YYYY HH24:MI:SS");
    }

    query_len = 4096 + strlen(cn) + strlen(limit) + 1;
    if(*offset) {
        query_len += 8 + strlen(offset);
    }
    query = malloc(query_len);
    strcpy(query, "SELECT * FROM (SELECT ");
    for(i=0; (i < num_column_lookup_table) && (i != DATUM_COL);i++) {
        if(i == CN_COL) {
            strcat(query, "cdr.cn");
        }
        else if(i == TIME_COL || i == CALLBACK_TIME) {
            strcat(query, "to_char(");
            strcat(query, column_lookup_table[i]);
            strcat(query, ", '");
            strcat(query, date_time);
            strcat(query, "') AS ");
            strcat(query, column_lookup_table[i]);
        }
        else {
            strcat(query, column_lookup_table[i]);
        }
        strcat(query, ",");
    }
    strcat(query, "local_stamp AS lstamp, cu.clear_report_time");
    strcat(query, " FROM cdrs cdr JOIN cdr_properties cdrp ON cdr.id=cdrp.cdr_id");
    strcat(query, " JOIN cdr_users cu ON cdr.cdr_user_id=cu.id WHERE cu.cn='");
    strcat(query, cn);
    strcat(query, "' AND cu.");
    strcat(query, reporting_cfg->pbx->identifier);
    strcat(query, "='");
    strcat(query, userpbx);
    strcat(query, "' AND cdr.transferred=0 AND cdr.forwarded=0 AND cdrp.cause_code != 26");
    strcat(query, " AND cdrp.cdrp_call_flow NOT LIKE '%{setup-from,,,},{rel-from,,,}%'");
    strftime(old_time, 11, "%Y-%m-%d", t);
    strcat(query, " AND local_stamp >= '");
    strcat(query, old_time);
    strcat(query, "'");
    strcat(query, " ORDER BY cdr.id DESC LIMIT ");
    strcat(query, limit);
    if(*offset) {
        strcat(query, " OFFSET ");
        strcat(query, offset);
    }
    strcat(query, ") AS temp");
    strcat(query, " WHERE (clear_report_time IS NULL OR lstamp > clear_report_time)");
    
    fwrite(CFG_CONN_DB, 1, strlen(CFG_CONN_DB), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(PG_CONN_STR, 1, strlen(PG_CONN_STR), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_MYPBX, 1, strlen(CFG_MYPBX), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite("1", 1, 1, fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_CN, 1, strlen(CFG_CN), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(cn, 1, strlen(cn), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_USER_GROUP, 1, strlen(CFG_USER_GROUP), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(reporting_cfg->pbx->identifier, 1, strlen(reporting_cfg->pbx->identifier), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_USER_PBX, 1, strlen(CFG_USER_PBX), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(userpbx, 1, strlen(userpbx), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_QUERY, 1, strlen(CFG_QUERY), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(query, 1, strlen(query), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_FORMAT, 1, strlen(CFG_FORMAT), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite("xsl", 1, strlen("xsl"), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite(CFG_DISPLAYED_DURATION, 1, strlen(CFG_DISPLAYED_DURATION), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    fwrite((char *)column_lookup_table[CALL_CONN_COL], 1, strlen(column_lookup_table[CALL_CONN_COL]), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    if(lang_str) {
        fwrite(CFG_LANGUAGE, 1, strlen(CFG_LANGUAGE), fp_tempfile);
        fwrite("\n", 1, 1, fp_tempfile);
        fwrite(lang_str, 1, strlen(lang_str), fp_tempfile);
        fwrite("\n", 1, 1, fp_tempfile);
    }
    fwrite(CFG_COL_NAMES, 1, strlen(CFG_COL_NAMES), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    sprintf(cols, "%d", num_column_lookup_table);
    fwrite(cols, 1, strlen(cols), fp_tempfile);
    fwrite("\n", 1, 1, fp_tempfile);
    for(i=0; i < num_column_lookup_table; i++) { 
        fwrite((char *)column_lookup_table[i], 1, strlen(column_lookup_table[i]), fp_tempfile);
        fwrite("\n", 1, 1, fp_tempfile);
    }
    fclose(fp_tempfile);
    
    // create the report
    sprintf(tmp_query_string, "cfg=%s", strstr(tmp_template, "mypbxcfg"));
    res = create_report(tmp_query_string, cgi, doc_root, bytes_sent, output_len);

    // delete the config file
    remove(tmp_template);
    if(offset) free(offset);
    if(limit) free(limit);
    if(query) free(query);
    if(date_time) free(date_time);
    if(lang_str) free(lang_str);
    return res;
}

int mypbx(char *query_string, FCGX_Request *cgi, char * doc_root, int *bytes_sent, int *output_len)
{
    CURL * conn = curl_easy_init();
    char * action = NULL;
    char * cn = NULL;
    char * sys = NULL;
    char * pbx = NULL;
    char * challenge = NULL;
    char * auth = NULL;
    char log_str[MAX_TRACE_LENGTH];
    const char * xsl = NULL;
    int log_len = 0;
    int ret = 0;
    config_t *reporting_cfg = NULL;
    reporting_cfg = read_config(NULL);
    if(reporting_cfg->error) {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx: couldn't read config: %s", reporting_cfg->error);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, "", cgi, bytes_sent, output_len);
        cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
        return ERROR_REPORT_PARSING_CFG_FILE;
    }

    action = get_url_param(conn, query_string, "action", "mypbx", LOG_REPORTING);
    cn = get_url_param(conn, query_string, "cn", "mypbx", LOG_REPORTING);
    pbx = get_url_param(conn, query_string, "location", "mypbx", LOG_REPORTING);
    sys = get_url_param(conn, query_string, "system", "mypbx", LOG_REPORTING);
    challenge = get_url_param(conn, query_string, "challenge", "mypbx", LOG_REPORTING);
    
    if(!cn || !*cn || !action || !*action || !pbx || !sys) {
        char * missing = "action";
        if(!cn || !*cn) missing = "cn";
        else if(!pbx) missing = "location";
        else if(!sys) missing = "system";
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx: %s GET parameter not set", missing);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, "", cgi, bytes_sent, output_len);
        cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
        return ERROR_MYPBX_PARAMETER_NOT_SET;
    }
    
    if(!strcasecmp(action, ACTION_REPORT)) {
        xsl = XSL_REPORT;
    }
    else if(!strcasecmp(action, ACTION_CALLBACK)) {
        xsl = XSL_CALLBACK;
    }
    else if(!strcasecmp(action, ACTION_MISSED)) {
        xsl = XSL_MISSED;
    }
    else if(!strcasecmp(action, ACTION_CLEAR)) {
        xsl = XSL_CLEAR;
    }
    else {
        log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx: action '%s' not defined", action);
        send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, "", cgi, bytes_sent, output_len);
        cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
        return ERROR_MYPBX_ACTION_NOT_DEFINED;
    }

    // just for testing!!!
    /*ret = report(conn, query_string, cn, cgi, doc_root, xsl, bytes_sent, output_len);
    cleanup(conn, action, cn, challenge, auth);
    return ret;*/

    if(!challenge || !*challenge) { // first authentication step
        // create a temporary file
        char tmp_template[256];
        snprintf(tmp_template, sizeof(tmp_template), "%s/apps/innovaphone-reporting/tmp/challenge_XXXXXX", doc_root);
        int tempfile = mkstemp(tmp_template);
        FILE * fp_tempfile = NULL;
        if(tempfile > 0 && (fp_tempfile = fdopen(tempfile, "w+"))) {
            fwrite(cn, 1, strlen(cn), fp_tempfile);
            fclose(fp_tempfile);
        }
        else {
            log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): temporary challenge file couldn't be created", action);
            send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
            cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
            return ERROR_MYPBX_CHALLENGE_NOT_CREATED;
        }
        char * file = strrchr(tmp_template, '/');
        file++;
        send_xml("challenge", file, strlen(file), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
    }
    else {  // second authentication step
        char file[256];
        snprintf(file, sizeof(file), "%s/apps/innovaphone-reporting/tmp/%s", doc_root, challenge);
        char * file_cn = read_file(file);
        if(!file_cn) {
            log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): challenge %s not valid anymore", action, challenge);
            send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
            cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
            return ERROR_MYPBX_CHALLENGE_NOT_VALID;
        }
        else {
            remove(file);
            if(strcasecmp(file_cn, cn)) {
                log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): GET CN %s does not match its challenge CN %s", action, cn, file_cn);
                send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
                cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
                free(file_cn);
                return ERROR_MYPBX_CN_NOT_MATCHING_CHALLENGE;
            }
            free(file_cn);
            auth = get_url_param(conn, query_string, "auth", "mypbx", LOG_REPORTING);
            if(!auth || !*auth) {
                log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): GET parameter auth not set", action);
                send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
                cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
                return ERROR_MYPBX_AUTH_NOT_SET;
            }
            char * lighttpd_digest = read_file("/home/lighttpd/lighttpd_htdigest.user");
            char * reporting_digest = read_file("/home/lighttpd/innovaphone-reporting-htdigest.user");
            char * digest = NULL;
            if(!reporting_digest) {
                digest = str_malloc(lighttpd_digest);
            }
            else {
                digest = malloc(strlen(lighttpd_digest) + strlen(reporting_digest) + 2);
                strcpy(digest, lighttpd_digest);
                if(digest[strlen(digest) - 1] != '\n') {
                    strcat(digest, "\n");
                }
                strcat(digest, reporting_digest);
                free(reporting_digest);
            }
            free(lighttpd_digest);

            // SHA1(MD5(user:Linux Web Server:password):cn:challenge) 
            char * line_start = digest;
            char * line_end = strchr(digest, '\n');
            int line_len = 0;
            int found = 0;
            unsigned char sha1[SHA_DIGEST_LENGTH + 1];
            char sha1_cmp[SHA_DIGEST_LENGTH * 2 + 1];
            while(1) {
                memset(sha1, 0, sizeof(sha1)*sizeof(unsigned char));
                memset(sha1_cmp, 0, sizeof(sha1_cmp)*sizeof(char));
                line_len = line_end ? (line_end - line_start) : strlen(line_start);
                char line[line_len + 1];
                strncpy(line, line_start, line_len);
                line[line_len] = 0;
                char * md5 = strrchr(line, ':');
                if(md5 && (++md5)) {
                    char to_sha1[strlen(md5) + strlen(cn) + strlen(challenge) + 3];
                    int to_sha1_len = sprintf(to_sha1, "%s:%s:%s", md5, cn, challenge);
                    SHA1((unsigned char *)to_sha1, to_sha1_len, (unsigned char*)sha1);
                    sprintf(sha1_cmp, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x", 
                            sha1[0], sha1[1], sha1[2], sha1[3], sha1[4], 
                            sha1[5], sha1[6], sha1[7], sha1[8], sha1[9], 
                            sha1[10], sha1[11], sha1[12], sha1[13], sha1[14], 
                            sha1[15], sha1[16], sha1[17], sha1[18], sha1[19]);
                    if(!strcmp(sha1_cmp, auth)) {
                        found = 1;
                        break;
                    }
                }

                if(!line_end) break;
                line_end++;
                if(!line_end) break;
                line_start = line_end;
                line_end = strchr(line_start, '\n');
            }
            free(digest);
            if(!found) {
                log_len = snprintf(log_str, MAX_TRACE_LENGTH, "mypbx (%s): received auth %s does not match %s", action, auth, sha1_cmp);
                send_xml("error", log_str, MIN(log_len, MAX_TRACE_LENGTH), TAG_AUTH, xsl, cgi, bytes_sent, output_len);
                cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
                return ERROR_MYPBX_WRONG_AUTH_RECV;
            }
            if(!strcasecmp(action, ACTION_REPORT)) {
                ret = report(conn, query_string, cn, sys, pbx, cgi, doc_root, xsl, bytes_sent, output_len, reporting_cfg);
                //if(ret != 0) ret += 10;
            }
            else if(!strcasecmp(action, ACTION_CALLBACK)) {
                ret = callback(conn, query_string, cgi, xsl, bytes_sent, output_len, reporting_cfg);
                //if(ret != 0) ret += 20;
            }
            else if(!strcasecmp(action, ACTION_MISSED)) {
                ret = missed_calls(conn, query_string, cn, sys, pbx, cgi, xsl, bytes_sent, output_len, reporting_cfg);
                //if(ret != 0) ret += 30;
            }
            else if(!strcasecmp(action, ACTION_CLEAR)) {
                ret = clear(conn, query_string, cn, sys, pbx, cgi, xsl, bytes_sent, output_len, reporting_cfg);
                //if(ret != 0) ret += 40;
            }
        }
    }

    cleanup(conn, action, cn, sys, pbx, challenge, auth, reporting_cfg);
    return ret;
}
