#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "inno-lib.h"
#include "libpq-fe.h"
#include "reporting_config.h"

#define PG_CONN_STR_REMOTE              "host = %s port = 5432 dbname = innovaphone-reporting user = innovaphone-reporting password = '%s' sslmode = require"
#define PG_CONN_STR_LOCAL               "host = 127.0.0.1 port = 5432 dbname = innovaphone-reporting user = innovaphone-reporting password = '' sslmode = disable"
#define QUERY_SELECT_MAX                "SELECT MAX(id) as id FROM cdrs"
#define QUERY_SELECT_MAX_DELETED        "SELECT MAX(id) as id FROM del_cdrs"
#define QUERY_LAST_ID                   "SELECT last_cdr_id FROM replication"
#define QUERY_LAST_DELETED_ID           "SELECT last_del_cdr_id FROM replication"
#define QUERY_INSERT_LAST_ID            "INSERT INTO replication VALUES (0, 0, date_trunc('second', localtimestamp))"
#define QUERY_UPDATE_REPLICATION        "UPDATE replication SET last_cdr_id=%d, last_replication_time=date_trunc('second', localtimestamp)"
#define QUERY_UPDATE_REPLICATION_DELETED "UPDATE replication SET last_del_cdr_id=%d, last_replication_time=date_trunc('second', localtimestamp)"
#define QUERY_CDR_GUIDS                 "SELECT id, cdr_guid FROM cdrs WHERE id BETWEEN %d AND %d ORDER BY id ASC"
#define QUERY_CDR_GUIDS_BEGIN           "SELECT c.cdr_guid AS cdr_guid, c.id AS cdr_id FROM cdrs c WHERE c.cdr_guid IN ("
#define QUERY_CDR_GUIDS_END             " AND NOT EXISTS (SELECT d.id FROM del_cdrs d WHERE d.cdr_guid=c.cdr_guid)"
#define QUERY_CDR_GUIDS_DELETE_BEGIN    "SELECT cdr_guid FROM del_cdrs WHERE cdr_guid IN ("
#define QUERY_CDR_GUIDS_DELETED         "SELECT id, cdr_guid FROM del_cdrs WHERE id BETWEEN %d AND %d ORDER BY id ASC"
#define QUERY_CDRS                      "SELECT * FROM cdrs WHERE id IN (%s) ORDER BY id ASC"
#define QUERY_EVENTS                    "SELECT * FROM events WHERE cdr_id IN (%s) ORDER BY cdr_id ASC, order_index ASC"
#define QUERY_GROUPS                    "SELECT * FROM groups WHERE cdr_id IN (%s) ORDER BY cdr_id ASC"
#define QUERY_PROPERTIES                "SELECT * FROM cdr_properties WHERE cdr_id IN (%s) ORDER BY cdr_id ASC"
#define QUERY_INSERT_RETURNING          " RETURNING id"
#define QUERY_INSERT_CDR_BEGIN          "INSERT INTO cdrs VALUES ("
#define QUERY_INSERT_EVENTS_BEGIN       "INSERT INTO events VALUES ("
#define QUERY_INSERT_PROP_BEGIN         "INSERT INTO cdr_properties VALUES ("
#define QUERY_INSERT_GROUPS_BEGIN       "INSERT INTO groups VALUES ("
#define QUERY_CDRS_DELETE               "DELETE FROM %s WHERE %s IN (%s)"
#define QUERY_INSERT_CDRS_DELETE_BEGIN  "INSERT INTO del_cdrs (cdr_guid) VALUES "

#define QUERY_DELETE_FILTERS            "DELETE FROM filter_times; DELETE FROM filter_conditions; DELETE FROM mail_filters; DELETE FROM filters;"
#define QUERY_MAX_FILTER_ID             "SELECT MAX(id) AS id, COUNT(id) AS count FROM filters WHERE id > %d ORDER BY id ASC LIMIT %d"
#define QUERY_FILTERS                   "SELECT * FROM filters WHERE id BETWEEN %d AND %d ORDER BY id ASC"
#define QUERY_FILTER_TIMES              "SELECT * FROM filter_times WHERE filter_id BETWEEN %d AND %d ORDER BY filter_id ASC"
#define QUERY_FILTER_CONDS              "SELECT * FROM filter_conditions WHERE filter_id BETWEEN %d AND %d ORDER BY filter_id ASC"
#define QUERY_INSERT_FILTERS_BEGIN      "INSERT INTO filters VALUES ("
#define QUERY_INSERT_FILTER_TIMES_BEGIN "INSERT INTO filter_times VALUES ("
#define QUERY_INSERT_FILTER_CONDS_BEGIN "INSERT INTO filter_conditions VALUES ("

#define QUERY_DELETE_MAILS              "DELETE FROM mail_filters; DELETE FROM mails;"
#define QUERY_MAX_MAIL_ID               "SELECT MAX(id) AS id, COUNT(id) AS count FROM mails WHERE id > %d ORDER BY id ASC LIMIT %d"
#define QUERY_MAILS                     "SELECT * FROM mails WHERE id BETWEEN %d AND %d ORDER BY id ASC"
#define QUERY_MAIL_FILTERS              "SELECT * FROM mail_filters WHERE mail_id BETWEEN %d AND %d ORDER BY mail_id ASC"
#define QUERY_INSERT_MAILS_BEGIN        "INSERT INTO mails VALUES ("
#define QUERY_INSERT_MAIL_FILTERS_BEGIN "INSERT INTO mail_filters VALUES ("

#define QUERY_DELETE_LDAP_SERVERS       "DELETE FROM ldap_servers;"
#define QUERY_MAX_LDAP_SERVER_ID        "SELECT MAX(id) AS id, COUNT(id) AS count FROM ldap_servers WHERE id > %d ORDER BY id ASC LIMIT %d"
#define QUERY_LDAP_SERVERS              "SELECT * FROM ldap_servers WHERE id BETWEEN %d AND %d ORDER BY id ASC"
#define QUERY_INSERT_LDAP_SERVERS_BEGIN "INSERT INTO ldap_servers VALUES ("

#define QUERY_DELETE_USERS              "DELETE FROM user_filters; DELETE FROM users;"
#define QUERY_MAX_USER_ID               "SELECT MAX(id) AS id, COUNT(id) AS count FROM users WHERE id > %d ORDER BY id ASC LIMIT %d"
#define QUERY_USERS                     "SELECT * FROM users WHERE id BETWEEN %d AND %d ORDER BY id ASC"
#define QUERY_USER_FILTERS              "SELECT * FROM user_filters WHERE user_id BETWEEN %d AND %d ORDER BY user_id ASC"
#define QUERY_INSERT_USERS_BEGIN        "INSERT INTO users VALUES ("
#define QUERY_INSERT_USER_FILTERS_BEGIN "INSERT INTO user_filters VALUES ("

#define ID_COUNT                   100
#define MAX_INT_LEN                 10
#define GUID_LEN                    32

struct copy_id {
    int old_id;
    int new_id;
};

void do_replication_log(int error, const char * format, ...) 
{
    // log time first
    time_t t;
    struct tm * tinfo;
    time(&t);
    tinfo = localtime(&t);
    char time_str[22];
    strftime(time_str, sizeof(time_str), "%Y-%m-%d %H:%M:%S ", tinfo);
    
    if(error) {
        printf("ERROR: %s", time_str);
    }
    else {
        printf("LOG: %s", time_str);
    }
    
    // log format with arguments
    va_list args;
    va_start (args, format);
    vprintf (format, args);
    va_end (args);
    printf("\r\n");
}

int status(int s, char * status_file)
{
    if(status_file) {
        FILE * f = fopen(status_file, "w");
        if(f) {
            fprintf(f, "%d", s);
            fclose(f);
        }
        else {
            do_replication_log(1, "Couldn't open status file %s", status_file);
        }
    }
    return s;
}

void cleanup(struct config_ * cfg, PGconn * pg_remote, char * conn_str, PGconn * pg_local, char ** guids) 
{
    int i = 0;
    if(cfg) cleanup_config(cfg);
    if(pg_remote) PQfinish(pg_remote);
    if(conn_str) free(conn_str);
    if(pg_local) PQfinish(pg_local);
    char * guid = NULL;
    for(i = 0; i < ID_COUNT; i++, guid = guids[i]) {
        if(guid) free(guid);
    }
}

PGconn * connect_to_db(char * conn_str, char * type) 
{
    PGconn * conn = PQconnectdb(conn_str);
    if (PQstatus(conn) != CONNECTION_OK) {
        do_replication_log(1, "Connection to %s database failed: %s", type, PQerrorMessage(conn));
        PQfinish(conn);
        return NULL;
    }

    if (PQsetClientEncoding(conn, "UTF-8") != 0) {
        do_replication_log(1, "Setting client encoding on %s database failed: %s", type, PQerrorMessage(conn));
        PQfinish(conn);
        return NULL;
    }
    return conn;
}

char * pg_escape_string(PGconn * conn, char * from, int is_null)
{
    if(is_null) return str_malloc("NULL");
    size_t to_len = strlen(from) * 2 + 1 + 2;   // + 2 because of the surrounding ''
    char * to = malloc(to_len) ;
    *to = '\'';
    int * err = NULL;
    PQescapeStringConn(conn, &to[1], (const char *)from, to_len, err);
    strcat(to, "'");
    return to;
}

int delete_rows(PGconn * pg_local, char * query_ids, int query_ids_len, char * table, char * column) 
{
    int query_len = query_ids_len + strlen(table) +  strlen(column) + strlen(QUERY_CDRS_DELETE) + 1;
    char * query = malloc(query_len);
    PGresult * res = NULL;
    sprintf(query, QUERY_CDRS_DELETE, table, column, query_ids);
    res = PQexec(pg_local, query);
    free(query);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_replication_log(1, "Deleting CDRs from %s failed: %s", table, PQerrorMessage(pg_local));
        PQclear(res);
        return 0;
    }
    PQclear(res);
    return 1;
}

int copy_rows(PGconn * pg_remote, PGconn * pg_local, char * query_ids, int query_ids_len, char * query_select, char * query_insert_begin, char * nextval, char * table, int id_from, int id_to, int is_cdr, struct copy_id * copy_ids)
{
    char ** cols;
    int i, j, k;
    char * query;
    int query_len;
    int nTuples = 0;
    int nFields = 0;
    int query_add_len = 0;
    int old_id = 0;
    PGresult * res = NULL;
    PGresult * res_insert = NULL;
    if(is_cdr) {
        for(i = 0; i < ID_COUNT; i++) {
            copy_ids[i].new_id = 0;
            copy_ids[i].old_id = 0;
        }
        query_add_len = strlen(QUERY_INSERT_RETURNING);
    }

    if(query_ids) {
        query = malloc(strlen(query_select) + query_ids_len + 1);
        sprintf(query, query_select, query_ids);
    }
    else {
        query = malloc(strlen(query_select) + (2 * MAX_INT_LEN) + 1);
        sprintf(query, query_select, id_from, id_to);
    }

    res = PQexec(pg_remote, query);
    free(query);
    if (PQresultStatus(res) != PGRES_TUPLES_OK) {
        do_replication_log(1, "Getting %s from remote server failed: %s", table, PQerrorMessage(pg_remote));
        PQclear(res);
        return 0;
    }
    nTuples = PQntuples(res);
    nFields = PQnfields(res);
    cols = malloc(nFields * sizeof(char *));
    for(i = 0; i < nTuples; i++) {
        memset(cols, 0, nFields * sizeof(char *));
        query_len = strlen(query_insert_begin) + 2;
        for(j = 0; j < nFields; j++) {  // ignore id
            if(nextval && !strcasecmp(PQfname(res, j), "id")) {
                cols[j] = str_malloc(nextval);
                if(is_cdr) {                                        // save remote id
                    copy_ids[i].old_id = atoi(PQgetvalue(res, i, j));
                }
            }
            else if(nextval && !strcasecmp(PQfname(res, j), "cdr_id")) {
                cols[j] = malloc(MAX_INT_LEN + 1);
                old_id = atoi(PQgetvalue(res, i, j));
                for(k = 0; k < ID_COUNT && copy_ids[k].old_id; k++) {
                    if(copy_ids[k].old_id == old_id) {
                        sprintf(cols[j], "%d", copy_ids[k].new_id);
                        break;
                    }
                }
            }
            else {
                cols[j] = pg_escape_string(pg_local, PQgetvalue(res, i, j), PQgetisnull(res, i, j));
            }
            query_len += strlen(cols[j]) + 1; // 1 because of ,
        }
        query = malloc(query_len + query_add_len);
        strcpy(query, query_insert_begin);
        for(j = 0; j < nFields && cols[j]; j++) {
            strcat(query, cols[j]);
            if((j + 1) < nFields) {
                strcat(query, ",");
            }
            free(cols[j]);
            cols[j] = NULL;
        }
        strcat(query, ")");
        if(is_cdr) {
            strcat(query, QUERY_INSERT_RETURNING);
        }
        //do_replication_log("%s", query);
        res_insert = PQexec(pg_local, query);
        free(query);
        if (PQresultStatus(res_insert) != PGRES_COMMAND_OK && PQresultStatus(res_insert) != PGRES_TUPLES_OK)
        {
            do_replication_log(1, "Inserting rows into %s failed: %s", table, PQerrorMessage(pg_local));
            free(cols);
            PQclear(res_insert);
            PQclear(res);
            return 0;
        } 
        if(is_cdr) {
            copy_ids[i].new_id = atoi(PQgetvalue(res_insert, 0, 0));
        }
        PQclear(res_insert);
    }
    free(cols);
    PQclear(res);
    return 1;
}

int main(int argc, char** argv) {
    struct config_ * cfg = NULL; 
    char * conn_str = NULL;
    PGresult * res = NULL;
    PGconn *pg_remote = NULL;
    PGconn *pg_local = NULL;
    char * query = NULL;
    char * query_ids = NULL;
    char * pg_value = NULL;
    struct copy_id copy_ids[ID_COUNT];
    unsigned int last_id = 0;
    unsigned int max_id = 0;
    unsigned int next_last_id = 0;
    int replicated_cdrs = 0;
    int deleted_cdrs = 0;
    int replicated_filters = 0;
    int replicated_mails = 0;
    int replicated_users = 0;
    int replicated_ldap_servers = 0;
    int nTuples = 0;
    char * guids[ID_COUNT];
    unsigned int ids[ID_COUNT];
    int i = 0, j = 0;
    int query_len = 0;
    int query_ids_len = 0;
    char * status_file = NULL;
    int insert_del_cdrs = 0;
    int fetched_del_cdrs = 0;
    trace_level = read_trace_level();
    
    if(argc == 1) {
        do_replication_log(1, "No status file given.");
    }
    else {
        status_file = argv[1];
    }

    memset(guids, 0, ID_COUNT * sizeof(char *));
    // read config file
    cfg = read_config(NULL);
    if(cfg->error) {
        do_replication_log(1, "%s\r\n", cfg->error);
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(1, status_file);
    }

    if(!strcasecmp(cfg->replication->mode, "off")) { // no replication
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(0, status_file);
    }
    else if(strcasecmp(cfg->replication->mode, "standby") && strcasecmp(cfg->replication->mode, "master")) {
        do_replication_log(1, "No valid replication mode configured: %s", cfg->replication->mode);
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(2, status_file);
    }
    else if(!cfg->replication->server || !cfg->replication->password) {
        do_replication_log(1, "No valid replication server/password configured");
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(3, status_file);
    }

    conn_str = malloc(strlen(PG_CONN_STR_REMOTE) + strlen(cfg->replication->password) + strlen(cfg->replication->server) + 1);
    sprintf(conn_str, PG_CONN_STR_REMOTE, cfg->replication->server, cfg->replication->password);

    if(!(pg_remote = connect_to_db(conn_str, "remote"))) {
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(4, status_file);
    }
    if(!(pg_local = connect_to_db(PG_CONN_STR_LOCAL, "local"))) {
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(5, status_file);
    }

    // now both dbs are connected
    // query max cdr id from replication server
    res = PQexec(pg_remote, QUERY_SELECT_MAX);
    if (PQresultStatus(res) != PGRES_TUPLES_OK || (nTuples = PQntuples(res)) == 0) {
        do_replication_log(1, "Fetching max cdr id from remote server failed: %s", PQerrorMessage(pg_remote));
        PQclear(res);
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(6, status_file);
    }
    pg_value = PQgetvalue(res, 0, 0);
    max_id = atoi(pg_value);
    PQclear(res);

    // query last cdr id from local server
    res = PQexec(pg_local, QUERY_LAST_ID);
    if (PQresultStatus(res) != PGRES_TUPLES_OK) {
        do_replication_log(1, "Fetching last cdr id from local server failed: %s", PQerrorMessage(pg_local));
        PQclear(res);
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(7, status_file);
    }
    nTuples = PQntuples(res);
    if(!nTuples) {
        PQclear(res);
        res = PQexec(pg_local, QUERY_INSERT_LAST_ID);
        if(PQresultStatus(res) != PGRES_COMMAND_OK) {
            do_replication_log(1, "Inserting last cdr id into replication table failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(8, status_file);
        }
        last_id = 0;
    }
    else {
        pg_value = PQgetvalue(res, 0, 0);
        last_id = atoi(pg_value);
    }
    PQclear(res);

    //do_replication_log("max=%d last=%d", max_id, last_id);

    do {
        next_last_id = last_id + ID_COUNT;
        if(next_last_id > max_id) {
            next_last_id = max_id;
        }

        query = malloc(strlen(QUERY_CDR_GUIDS) + MAX_INT_LEN * 2 + 1);
        sprintf(query, QUERY_CDR_GUIDS, last_id + 1, next_last_id);
        res = PQexec(pg_remote, query);
        free(query);
        if (PQresultStatus(res) != PGRES_TUPLES_OK) {
            do_replication_log(1, "Fetching cdr guids from remote server between %d and %d failed: %s", last_id + 1, next_last_id, PQerrorMessage(pg_remote));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(9, status_file);
        }
        nTuples = PQntuples(res);
        if(nTuples) {   // are there any cdrs for these ids?
            query_len = strlen(QUERY_CDR_GUIDS_BEGIN) + strlen(QUERY_CDR_GUIDS_END) + 2;  // 2 because of ) and \0
            for(i = 0; i < nTuples; i++) {
                guids[i] = str_malloc(PQgetvalue(res, i, 1));
                ids[i] = atoi(PQgetvalue(res, i, 0));
                query_len += strlen(guids[i]) + 3;  // 3 because of '',
                //do_replication_log("tuple %d guid %s id %d", i, guids[i], ids[i]);
            }
            PQclear(res);
            query = malloc(query_len);
            strcpy(query, QUERY_CDR_GUIDS_BEGIN);
            for(i = 0; i < nTuples; i++) {
                strcat(query, "'");
                strcat(query, guids[i]);
                strcat(query, "',");
            }
            query_len = strlen(query);
            query[query_len - 1] = ')';
            query[query_len] = 0;
            strcat(query, QUERY_CDR_GUIDS_END);
            //do_replication_log("%s", query);

            res = PQexec(pg_local, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Checking cdr guids from local server failed: %s", PQerrorMessage(pg_local));
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(10, status_file);
            }
            nTuples = PQntuples(res);
            for(i = 0; i < nTuples; i++) {
                pg_value = PQgetvalue(res, i, 0);
                for(j = 0; j < ID_COUNT; j++) {    // check if guid already exists in DB
                    if(guids[j] && !strcasecmp(guids[j], pg_value)) {
                        free(guids[j]);
                        guids[j] = NULL;
                        break;
                    }
                }
            }
            PQclear(res);
            // now fetch all stuff from the remaining guids
            query_ids = malloc(ID_COUNT * (MAX_INT_LEN + 1) + 1);
            *query_ids = 0;
            query_ids_len = 0;
            for(i = 0; i < ID_COUNT; i++) {
                if(!guids[i]) continue;
                query_ids_len += sprintf(&query_ids[query_ids_len], "%d,", ids[i]);
                free(guids[i]);
                guids[i] = NULL;
                replicated_cdrs++;
            }
        }
        else {
            PQclear(res);
            query_ids_len = 0;
            query_ids = NULL;
        }
        last_id = next_last_id;
        res = PQexec(pg_local, "BEGIN");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction begin failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            if(query_ids) free(query_ids);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(11, status_file);
        } 
        PQclear(res);
        // no ids to insert found?
        if(query_ids_len > 0) {
            query_ids[query_ids_len - 1] = 0;
        
            //do_replication_log("query ids = '%s'", query_ids);
            
            if(!copy_rows(pg_remote, pg_local, query_ids, query_ids_len, QUERY_CDRS, QUERY_INSERT_CDR_BEGIN, "nextval('cdr_id_seq'::regclass)", "cdrs", 0, 0, 1, copy_ids) ||
                !copy_rows(pg_remote, pg_local, query_ids, query_ids_len, QUERY_EVENTS, QUERY_INSERT_EVENTS_BEGIN, "nextval('events_id_seq'::regclass)", "events", 0, 0, 0, copy_ids) ||
                !copy_rows(pg_remote, pg_local, query_ids, query_ids_len, QUERY_PROPERTIES, QUERY_INSERT_PROP_BEGIN, "nextval('cdr_properties_id_seq'::regclass)", "cdr_properties", 0, 0, 0, copy_ids) ||
                !copy_rows(pg_remote, pg_local, query_ids, query_ids_len, QUERY_GROUPS, QUERY_INSERT_GROUPS_BEGIN, "nextval('groups_id_seq'::regclass)", "groups", 0, 0, 0, copy_ids)) {
                free(query_ids);
                res = PQexec(pg_local, "ROLLBACK");
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(12, status_file);
            }
        }
        if(query_ids) free(query_ids);
        query = malloc(strlen(QUERY_UPDATE_REPLICATION) + MAX_INT_LEN + 1);
        sprintf(query, QUERY_UPDATE_REPLICATION, last_id);
        res = PQexec(pg_local, query);
        free(query);
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Updating replication last cdr id to %d failed: %s", last_id, PQerrorMessage(pg_local));
            PQclear(res);
            res = PQexec(pg_local, "ROLLBACK");
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(13, status_file);
        } 
        PQclear(res);

        res = PQexec(pg_local, "COMMIT");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction commit failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(14, status_file);
        } 
        PQclear(res);
    } while(last_id < max_id);
    do_replication_log(0, "Replication of %d CDRs finished.", replicated_cdrs);
    // "replicate" all deleted CDRs...
     // query max deleted cdr id from replication server
    res = PQexec(pg_remote, QUERY_SELECT_MAX_DELETED);
    if (PQresultStatus(res) != PGRES_TUPLES_OK || (nTuples = PQntuples(res)) == 0) {
        do_replication_log(1, "Fetching max deleted cdr id from remote server failed: %s", PQerrorMessage(pg_remote));
        PQclear(res);
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(15, status_file);
    }
    pg_value = PQgetvalue(res, 0, 0);
    max_id = atoi(pg_value);
    PQclear(res);
    // query last deleted cdr id from replication server
    res = PQexec(pg_local, QUERY_LAST_DELETED_ID);
    if (PQresultStatus(res) != PGRES_TUPLES_OK) {
        do_replication_log(1, "Fetching last deleted cdr id from local server failed: %s", PQerrorMessage(pg_local));
        PQclear(res);
        cleanup(cfg, pg_remote, conn_str, pg_local, guids);
        return status(16, status_file);
    }
    pg_value = PQgetvalue(res, 0, 0);
    last_id = atoi(pg_value);
    PQclear(res);
    do {
        next_last_id = last_id + ID_COUNT;
        if(next_last_id > max_id) {
            next_last_id = max_id;
        }

        query = malloc(strlen(QUERY_CDR_GUIDS_DELETED) + MAX_INT_LEN * 2 + 1);
        sprintf(query, QUERY_CDR_GUIDS_DELETED, last_id + 1, next_last_id);
        res = PQexec(pg_remote, query);
        free(query);
        if (PQresultStatus(res) != PGRES_TUPLES_OK) {
            do_replication_log(1, "Fetching deleted cdr guids from remote server between %d and %d failed: %s", last_id + 1, next_last_id, PQerrorMessage(pg_remote));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(17, status_file);
        }
        nTuples = fetched_del_cdrs = PQntuples(res);
        if(nTuples) {   // are there any cdrs for these ids?
            query_len = strlen(QUERY_CDR_GUIDS_BEGIN) + 2;  // 2 because of ) and \0
            query_len += (nTuples * (GUID_LEN + 3)); // 3 because of '',
            query = malloc(query_len);
            strcpy(query, QUERY_CDR_GUIDS_BEGIN);
            for(i = 0; i < nTuples; i++) {
                strcat(query, "'");
                strcat(query, PQgetvalue(res, i, 1));
                strcat(query, "',");
                guids[i] = str_malloc(PQgetvalue(res, i, 1));
            }
            PQclear(res);
            query_len = strlen(query);
            query[query_len - 1] = ')';
            query[query_len] = 0;
            //do_replication_log("%s", query);

            res = PQexec(pg_local, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Checking deleted cdr guids from local server failed: %s", PQerrorMessage(pg_local));
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(18, status_file);
            }
            nTuples = PQntuples(res);
            deleted_cdrs += nTuples;
            query_ids = malloc(nTuples * (MAX_INT_LEN + 1) + 1);
            *query_ids = 0;
            query_ids_len = 0;
            for(i = 0; i < nTuples; i++) {
                query_ids_len += sprintf(&query_ids[query_ids_len], "%d,", atoi(PQgetvalue(res, i, 1)));
            }
            PQclear(res);
        }
        else {
            PQclear(res);
            query_ids_len = 0;
            query_ids = NULL;
        }
        last_id = next_last_id;
        res = PQexec(pg_local, "BEGIN");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction for deleted CDRs begin failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            if(query_ids) free(query_ids);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(19, status_file);
        } 
        PQclear(res);
        // no ids to insert found?
        if(query_ids_len > 0) {
            query_ids[query_ids_len - 1] = 0;
            //do_replication_log("query ids = '%s'", query_ids);
            
            if(!delete_rows(pg_local, query_ids, query_ids_len, "events", "cdr_id") ||
                !delete_rows(pg_local, query_ids, query_ids_len, "cdr_properties", "cdr_id") ||
                !delete_rows(pg_local, query_ids, query_ids_len, "groups", "cdr_id") ||
                !delete_rows(pg_local, query_ids, query_ids_len, "cdrs", "id")) {
                free(query_ids);
                res = PQexec(pg_local, "ROLLBACK");
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(20, status_file);
            }
        }
        if(query_ids) free(query_ids);
        if(fetched_del_cdrs > 0) {
            query = malloc(strlen(QUERY_CDR_GUIDS_DELETE_BEGIN) + (fetched_del_cdrs * (GUID_LEN + 3)) + 2);
            strcpy(query, QUERY_CDR_GUIDS_DELETE_BEGIN);
            for(i = 0; i < fetched_del_cdrs; i++) {
                if(guids[i]) {
                    strcat(query, "'");
                    strcat(query, guids[i]);
                    strcat(query, "',");
                }
            }
            query_len = strlen(query);
            query[query_len - 1] = ')';
            query[query_len] = 0;
            //do_replication_log("%s", query);
            res = PQexec(pg_local, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Checking deleted cdr guids from local server failed: %s", PQerrorMessage(pg_local));
                PQclear(res);
                res = PQexec(pg_local, "ROLLBACK");
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(21, status_file);
            }
            nTuples = PQntuples(res);
            for(i = 0; i < nTuples; i++) {
                pg_value = PQgetvalue(res, i, 0);
                for(j = 0; j < fetched_del_cdrs; j++) {    // check if guid already exists in DB
                    if(guids[j] && !strcasecmp(guids[j], pg_value)) {
                        free(guids[j]);
                        guids[j] = NULL;
                        break;
                    }
                }
            }
            for(i = 0; i < fetched_del_cdrs; i++) {
                if(guids[i]) {
                    insert_del_cdrs = 1;
                    break;
                }
            }
            if(insert_del_cdrs) {
                query = malloc(strlen(QUERY_INSERT_CDRS_DELETE_BEGIN) + (fetched_del_cdrs * (GUID_LEN + 5)) + 2); // 2 because of ) \0
                *query = 0;
                strcat(query, QUERY_INSERT_CDRS_DELETE_BEGIN);
                for(i = 0; i < fetched_del_cdrs; i++) {
                    if(guids[i]) {
                        strcat(query, "('");
                        strcat(query, guids[i]);
                        strcat(query, "'),");
                        free(guids[i]);
                        guids[i] = NULL;
                    }
                }
                query_len = strlen(query);
                query[query_len - 1] = 0;
                res = PQexec(pg_local, query);
                free(query);
                if (PQresultStatus(res) != PGRES_COMMAND_OK)
                {
                    do_replication_log(1, "Inserting deleted cdr guids failed: %s", PQerrorMessage(pg_local));
                    PQclear(res);
                    res = PQexec(pg_local, "ROLLBACK");
                    PQclear(res);
                    cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                    return status(22, status_file);
                } 
                PQclear(res);
            }
        }

        query = malloc(strlen(QUERY_UPDATE_REPLICATION_DELETED) + MAX_INT_LEN + 1);
        sprintf(query, QUERY_UPDATE_REPLICATION_DELETED, last_id);
        res = PQexec(pg_local, query);
        free(query);
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Updating replication last deleted cdr id to %d failed: %s", last_id, PQerrorMessage(pg_local));
            PQclear(res);
            res = PQexec(pg_local, "ROLLBACK");
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(23, status_file);
        } 
        PQclear(res);

        res = PQexec(pg_local, "COMMIT");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction commit failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(24, status_file);
        } 
        PQclear(res);
    } while(last_id < max_id);
    do_replication_log(0, "Replication of %d deleted CDRs finished.", deleted_cdrs);
    // replicate all filters/mails/users/ldaps from master if mode is standby
    if(!strcasecmp(cfg->replication->mode, "standby")) {
        res = PQexec(pg_local, "BEGIN");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction begin for filter replication failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(25, status_file);
        } 
        PQclear(res);
        // delete all current mails (before deleting the users/filters, as users are referenced here)
        res = PQexec(pg_local, QUERY_DELETE_MAILS);
        if (PQresultStatus(res) != PGRES_COMMAND_OK) {
            do_replication_log(1, "Deleting report mails on standby failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(31, status_file);
        }
        PQclear(res);
        // delete all current users (before deleting the filters)
        res = PQexec(pg_local, QUERY_DELETE_USERS);
        if (PQresultStatus(res) != PGRES_COMMAND_OK) {
            do_replication_log(1, "Deleting users on standby failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(31, status_file);
        }
        PQclear(res);
        // delete all current filters
        res = PQexec(pg_local, QUERY_DELETE_FILTERS);
        if (PQresultStatus(res) != PGRES_COMMAND_OK) {
            do_replication_log(1, "Deleting filters on standby failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(26, status_file);
        }
        PQclear(res);
        // fetch filters from master at first
        last_id = 0;
        do {
            query = malloc(strlen(QUERY_MAX_FILTER_ID) + (MAX_INT_LEN * 2) + 1);
            sprintf(query, QUERY_MAX_FILTER_ID, last_id, ID_COUNT);
            res = PQexec(pg_remote, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Fetching max filter id from remote server failed: %s", PQerrorMessage(pg_remote));
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(27, status_file);
            }
            next_last_id = atoi(PQgetvalue(res, 0, 0)); // MAX(id)
            nTuples = atoi(PQgetvalue(res, 0, 1));      // COUNT(id)
            replicated_filters += nTuples;
            if(nTuples) {
                if(!copy_rows(pg_remote, pg_local, NULL, 0, QUERY_FILTERS, QUERY_INSERT_FILTERS_BEGIN, NULL, "filters", last_id, next_last_id, 0, copy_ids) ||
                    !copy_rows(pg_remote, pg_local, NULL, 0, QUERY_FILTER_TIMES, QUERY_INSERT_FILTER_TIMES_BEGIN, NULL, "filter_times", last_id, next_last_id, 0, copy_ids) ||
                    !copy_rows(pg_remote, pg_local, NULL, 0, QUERY_FILTER_CONDS, QUERY_INSERT_FILTER_CONDS_BEGIN, NULL, "filter_conditions", last_id, next_last_id, 0, copy_ids)) {
                    res = PQexec(pg_local, "ROLLBACK");
                    PQclear(res);
                    cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                    return status(28, status_file);
                }
            }
            last_id = next_last_id;
        } while(nTuples == ID_COUNT);
        // fetch users next, as mail reports reference the users
        last_id = 0;
        do {
            query = malloc(strlen(QUERY_MAX_USER_ID) + (MAX_INT_LEN * 2) + 1);
            sprintf(query, QUERY_MAX_USER_ID, last_id, ID_COUNT);
            res = PQexec(pg_remote, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Fetching max user id from remote server failed: %s", PQerrorMessage(pg_remote));
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(32, status_file);
            }
            next_last_id = atoi(PQgetvalue(res, 0, 0)); // MAX(id)
            nTuples = atoi(PQgetvalue(res, 0, 1));      // COUNT(id)
            replicated_users += nTuples;
            if(nTuples) {
                if(!copy_rows(pg_remote, pg_local, NULL, 0, QUERY_USERS, QUERY_INSERT_USERS_BEGIN, NULL, "users", last_id, next_last_id, 0, copy_ids) ||
                    !copy_rows(pg_remote, pg_local, NULL, 0, QUERY_USER_FILTERS, QUERY_INSERT_USER_FILTERS_BEGIN, NULL, "user_filters", last_id, next_last_id, 0, copy_ids)) {
                    res = PQexec(pg_local, "ROLLBACK");
                    PQclear(res);
                    cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                    return status(33, status_file);
                }
            }
            last_id = next_last_id;
        } while(nTuples == ID_COUNT);
        // then fetch the mail reports
        last_id = 0;
        do {
            query = malloc(strlen(QUERY_MAX_MAIL_ID) + (MAX_INT_LEN * 2) + 1);
            sprintf(query, QUERY_MAX_MAIL_ID, last_id, ID_COUNT);
            res = PQexec(pg_remote, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Fetching max report mail id from remote server failed: %s", PQerrorMessage(pg_remote));
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(32, status_file);
            }
            next_last_id = atoi(PQgetvalue(res, 0, 0)); // MAX(id)
            nTuples = atoi(PQgetvalue(res, 0, 1));      // COUNT(id)
            replicated_mails += nTuples;
            if(nTuples) {
                if(!copy_rows(pg_remote, pg_local, NULL, 0, QUERY_MAILS, QUERY_INSERT_MAILS_BEGIN, NULL, "mails", last_id, next_last_id, 0, copy_ids) ||
                    !copy_rows(pg_remote, pg_local, NULL, 0, QUERY_MAIL_FILTERS, QUERY_INSERT_MAIL_FILTERS_BEGIN, NULL, "mail_filters", last_id, next_last_id, 0, copy_ids)) {
                    res = PQexec(pg_local, "ROLLBACK");
                    PQclear(res);
                    cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                    return status(33, status_file);
                }
            }
            last_id = next_last_id;
        } while(nTuples == ID_COUNT);
        res = PQexec(pg_local, "COMMIT");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction commit for filter/user/report mail replication failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(34, status_file);
        } 
        PQclear(res);
        do_replication_log(0, "Replication of %d filters finished.", replicated_filters);
        do_replication_log(0, "Replication of %d users finished.", replicated_users);
        do_replication_log(0, "Replication of %d report mails finished.", replicated_mails);

        // replicate ldap servers
        res = PQexec(pg_local, "BEGIN");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction begin for ldap server replication failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(35, status_file);
        } 
        PQclear(res);
        // delete all current ldap servers
        res = PQexec(pg_local, QUERY_DELETE_LDAP_SERVERS);
        if (PQresultStatus(res) != PGRES_COMMAND_OK) {
            do_replication_log(1, "Deleting ldap servers on standby failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(36, status_file);
        }
        PQclear(res);
        last_id = 0;
        do {
            query = malloc(strlen(QUERY_MAX_LDAP_SERVER_ID) + (MAX_INT_LEN * 2) + 1);
            sprintf(query, QUERY_MAX_LDAP_SERVER_ID, last_id, ID_COUNT);
            res = PQexec(pg_remote, query);
            free(query);
            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_replication_log(1, "Fetching max ldap server id from remote server failed: %s", PQerrorMessage(pg_remote));
                PQclear(res);
                cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                return status(37, status_file);
            }
            next_last_id = atoi(PQgetvalue(res, 0, 0)); // MAX(id)
            nTuples = atoi(PQgetvalue(res, 0, 1));      // COUNT(id)
            replicated_ldap_servers += nTuples;
            if(nTuples) {
                if(!copy_rows(pg_remote, pg_local, NULL, 0, QUERY_LDAP_SERVERS, QUERY_INSERT_LDAP_SERVERS_BEGIN, NULL, "ldap_servers", last_id, next_last_id, 0, copy_ids)) {
                    res = PQexec(pg_local, "ROLLBACK");
                    PQclear(res);
                    cleanup(cfg, pg_remote, conn_str, pg_local, guids);
                    return status(38, status_file);
                }
            }
            last_id = next_last_id;
        } while(nTuples == ID_COUNT);
        res = PQexec(pg_local, "COMMIT");
        if (PQresultStatus(res) != PGRES_COMMAND_OK)
        {
            do_replication_log(1, "Transaction commit for ldap server replication failed: %s", PQerrorMessage(pg_local));
            PQclear(res);
            cleanup(cfg, pg_remote, conn_str, pg_local, guids);
            return status(39, status_file);
        } 
        PQclear(res);
        do_replication_log(0, "Replication of %d ldap servers finished.", replicated_ldap_servers);
    }
    cleanup(cfg, pg_remote, conn_str, pg_local, guids);
    return status(0, status_file);
}
